/*************************************************************************
 * Clus - Software for Predictive Clustering                             *
 * Copyright (C) 2007                                                    *
 *    Katholieke Universiteit Leuven, Leuven, Belgium                    *
 *    Jozef Stefan Institute, Ljubljana, Slovenia                        *
 *                                                                       *
 * This program is free software: you can redistribute it and/or modify  *
 * it under the terms of the GNU General Public License as published by  *
 * the Free Software Foundation, either version 3 of the License, or     *
 * (at your option) any later version.                                   *
 *                                                                       *
 * This program is distributed in the hope that it will be useful,       *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 * GNU General Public License for more details.                          *
 *                                                                       *
 * You should have received a copy of the GNU General Public License     *
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 *                                                                       *
 * Contact information: <http://www.cs.kuleuven.be/~dtai/clus/>.         *
 *************************************************************************/

package clus.data.rows;

import jeans.util.sort.*;
import jeans.util.compound.*;

import java.util.*;
import java.io.*;

import clus.util.*;
import clus.algo.tdidt.ClusNode;
import clus.data.ClusData;
import clus.data.io.ClusReader;
import clus.data.type.*;
import clus.model.test.*;
import clus.error.*;
import clus.selection.*;
import clus.statistic.*;

/**
 * Multiple rows (tuples) of data.
 * One row (DataTuple) is one instance of the data with target and description attributes.
 */
public class RowData extends ClusData implements MSortable {

	public int m_Index;
	public ClusSchema m_Schema;
	public DataTuple[] m_Data;

	public RowData(ClusSchema schema) {
		m_Schema = schema;
	}

	public RowData(ClusSchema schema, int size) {
		m_Schema = schema;
		resizeEmpty(size);
	}

	public RowData(RowData data) {
		this(data.m_Data, data.getNbRows());
		m_Schema = data.m_Schema;
	}

	public RowData(Object[] data, int size) {
		m_Data = new DataTuple[size];
		System.arraycopy(data, 0, m_Data, 0, size);
		setNbRows(size);
	}

	public RowData(ArrayList list, ClusSchema schema) {
		m_Schema = schema;
		m_Data = new DataTuple[list.size()];
		for (int i = 0; i < list.size(); i++) {
			m_Data[i] = (DataTuple)list.get(i);
		}
		setNbRows(list.size());
	}

	public String toString(){
		return toString("");
	}

	public String toString(String prefix){
		StringBuffer sb = new StringBuffer();
		for (int i = 0; i < getNbRows(); i++) {
			sb.append(prefix);
			sb.append(getTuple(i).toString()+"\n");
		}
		return sb.toString();
	}

	public String getSummary(){
		return getSummary("");
	}

	public String getSummary(String prefix){
		StringBuffer sb = new StringBuffer();
		float[] avg,min,max,stddev;
		DataTuple temp = getTuple(0);
		int nda= temp.getSchema().getNbNumericDescriptiveAttributes();
		avg = new float[nda];
		min = new float[nda];
		max = new float[nda];
		stddev = new float[nda];
		Arrays.fill(avg, 0);
		Arrays.fill(stddev, 0);
		Arrays.fill(min, Float.MAX_VALUE);
		Arrays.fill(max, Float.MIN_VALUE);
		int nbrows = getNbRows();
		for (int i = 0; i < nbrows ; i++) {
			temp = getTuple(i);
			ClusSchema schema = temp.getSchema();
			for (int j = 0; j < schema.getNbNumericDescriptiveAttributes(); j++) {
				ClusAttrType type = schema.getNumericAttrUse(ClusAttrType.ATTR_USE_DESCRIPTIVE)[j];
				float tmpvalue = type.getNumeric(temp);
				if (tmpvalue>max[j]){
					max[j]=tmpvalue;
				}
				if (tmpvalue<min[j]){
					min[j]=tmpvalue;
				}
				avg[j]+=tmpvalue;
				stddev[j]+=tmpvalue*tmpvalue;
			}
		}
		for (int i=0;i<nda;i++){
			avg[i]/=nbrows;
			stddev[i]=(stddev[i]-nbrows*avg[i]*avg[i])/nbrows;
			stddev[i] =(float) Math.sqrt(stddev[i]);
			min[i]=Math.round(min[i]*100)/100.0f;
			max[i]=Math.round(max[i]*100)/100.0f;
			avg[i]=Math.round(avg[i]*100)/100.0f;
			stddev[i]=Math.round(stddev[i]*100)/100.0f;
		}

		sb.append(prefix+"Min: "+Arrays.toString(min)+"\n");
		sb.append(prefix+"Max: "+Arrays.toString(max)+"\n");
		sb.append(prefix+"Avg: "+Arrays.toString(avg)+"\n");
		sb.append(prefix+"StdDev: "+Arrays.toString(stddev)+"\n");
		return sb.toString();
	}

	public ArrayList toArrayList() {
		ArrayList array = new ArrayList();
		addTo(array);
		return array;
	}

	public void addTo(ArrayList array) {
		for (int i = 0; i < getNbRows(); i++) {
			array.add(getTuple(i));
		}
	}

	public DataTuple createTuple() {
		return new DataTuple(m_Schema);
	}

	public static RowData readData(String fname, ClusSchema schema) throws ClusException, IOException {
		schema.addIndices(ClusSchema.ROWS);
		ClusReader reader = new ClusReader(fname, schema.getSettings());
		RowData data = schema.createNormalView().readData(reader, schema);
		reader.close();
		return data;
	}

	public ClusData cloneData() {
		RowData res = new RowData(m_Schema, m_NbRows);
		System.arraycopy(m_Data, 0, res.m_Data, 0, m_NbRows);
		return res;
	}
	
	public RowData shallowCloneData() {
		RowData res = new RowData(m_Schema, m_NbRows);
		for (int i = 0; i < m_NbRows; i++) {
			res.setTuple(m_Data[i].cloneTuple(), i);
		}
		return res;
	}	

	public ClusData deepCloneData() {
		RowData res = new RowData(m_Schema, m_NbRows);
		for (int i = 0; i < m_NbRows; i++) {
			res.setTuple(m_Data[i].deepCloneTuple(), i);
		}
		return res;
	}

	public final ClusSchema getSchema() {
		return m_Schema;
	}

	public void setSchema(ClusSchema schema) {
		m_Schema = schema;
	}

	public void sortSparse(NumericAttrType at) {
		int nbmiss = 0, nbzero = 0, nbother = 0;
		DataTuple[] missing = new DataTuple[m_NbRows];
		DataTuple[] zero = new DataTuple[m_NbRows];
		DoubleObject[] other = new DoubleObject[m_NbRows];
		for (int i = 0; i < m_NbRows; i++) {
			float data = at.getNumeric(m_Data[i]);
			if (data == 0.0f) {
				zero[nbzero++] = m_Data[i];
			} else if (data == NumericAttrType.MISSING) {
				missing[nbmiss++] = m_Data[i];
			} else if (data > 0.0f) {
				other[nbother++] = new DoubleObject(data, m_Data[i]);
			} else {
				System.err.println("Sparse attribute has negative value!");
				System.exit(-1);
			}
		}
		MSorter.quickSort(new MySortableArray(other), 0, nbother);
		int pos = 0;
		for (int i = 0; i < nbmiss; i++) {
			m_Data[pos++] = missing[i];
		}
		for (int i = 0; i < nbother; i++) {
			m_Data[pos++] = (DataTuple)other[i].getObject();
		}
		for (int i = 0; i < nbzero; i++) {
			m_Data[pos++] = zero[i];
		}
	}

	public void sort(NumericAttrType at) {
		m_Index = at.getArrayIndex();
		MSorter.quickSort(this, 0, m_NbRows);
	}

	public float getfloat(int i) {
		return m_Data[i].getfloatVal(m_Index);
	}

	public boolean compare(int i, int j) {
		return m_Data[i].getfloatVal(m_Index) < m_Data[j].getfloatVal(m_Index);
	}

	public void swap(int i, int j) {
		DataTuple temp = m_Data[i];
		m_Data[i] = m_Data[j];
		m_Data[j] = temp;
	}

	public DataTuple findTupleByKey(String key_value) {
		ClusAttrType[] key = getSchema().getAllAttrUse(ClusAttrType.ATTR_USE_KEY);
		if (key.length > 0) {
			ClusAttrType key_attr = key[0];
			for (int i = 0; i < getNbRows(); i++) {
				DataTuple tuple = getTuple(i);
				if (key_attr.getString(tuple).equals(key_value)) return tuple;
			}
		}
		return null;
	}

	// Does not change original distribution
	public ClusData selectFrom(ClusSelection sel) {
		int nbsel = sel.getNbSelected();
		RowData res = new RowData(m_Schema, nbsel);
		if (sel.supportsReplacement()) {
			for (int i = 0; i < nbsel; i++) {
				res.setTuple(m_Data[sel.getIndex(i)], i);
			}
		} else {
			int s_subset = 0;
			for (int i = 0; i < m_NbRows; i++) {
				if (sel.isSelected(i)) res.setTuple(m_Data[i], s_subset++);
			}
		}
		return res;
	}

	public ClusData select(ClusSelection sel) {
		int s_data = 0;
		int s_subset = 0;
		DataTuple[] old = m_Data;
		int nbsel = sel.getNbSelected();
		m_Data = new DataTuple[m_NbRows - nbsel];
		RowData res = new RowData(m_Schema, nbsel);
		for (int i = 0; i < m_NbRows; i++) {
			if (sel.isSelected(i)) res.setTuple(old[i], s_subset++);
			else setTuple(old[i], s_data++);
		}
		m_NbRows -= nbsel;
		return res;
	}

	public void update(ClusSelection sel) {
		int s_data = 0;
		DataTuple[] old = m_Data;
		int nbsel = sel.getNbSelected();
		m_Data = new DataTuple[nbsel];
		for (int i = 0; i < m_NbRows; i++) {
			if (sel.isSelected(i)) {
				DataTuple nt = old[i].multiplyWeight(sel.getWeight(i));
				setTuple(nt, s_data++);
			}
		}
		m_NbRows = nbsel;
	}

	public final float getSumWeights() {
		float sum = 0.0f;
		for (int i = 0; i < m_NbRows; i++) {
			sum += m_Data[i].getWeight();
		}
		return sum;
	}

	public final boolean containsFold(DataTuple tuple, int[] folds) {
		for (int i = 0; i < folds.length; i++) {
			if (tuple.m_Folds[folds[i]] > 0) return true;
		}
		return false;
	}

	public final void optimize2(int[] folds) {
		int nbsel = 0;
		int s_data = 0;
		for (int i = 0; i < m_NbRows; i++) {
			if (containsFold(m_Data[i], folds)) nbsel++;
		}
		DataTuple[] old = m_Data;
		m_Data = new DataTuple[nbsel];
		for (int i = 0; i < m_NbRows; i++) {
			if (containsFold(old[i], folds)) {
				setTuple(old[i], s_data++);
			}
		}
		m_NbRows = nbsel;
	}

	public void insert(ClusData data, ClusSelection sel) {
		int s_data = 0;
		int s_subset = 0;
		DataTuple[] old = m_Data;
		RowData other = (RowData)data;
		resizeEmpty(m_NbRows + sel.getNbSelected());
		for (int i = 0; i < m_NbRows; i++) {
			if (sel.isSelected(i)) setTuple(other.getTuple(s_subset++), i);
			else setTuple(old[s_data++], i);
		}
	}

	public final RowData getFoldData(int fold) {
		int idx = 0;
		int nbsel = 0;
		// Count examples for fold
		for (int i = 0; i < m_NbRows; i++)
			if (m_Data[i].getIndex() != fold) nbsel++;
		// Select examples
		RowData res = new RowData(m_Schema, nbsel);
		for (int i = 0; i < m_NbRows; i++) {
			DataTuple tuple = m_Data[i];
			if (tuple.getIndex() != fold) res.setTuple(tuple, idx++);
		}
		return res;
	}

	/**
	 * Only used in efficient XVal code
	 * TODO Could be a bug: changeWeight -> multiplyWeight
	 * @return
	 */ 
	public final RowData getFoldData2(int fold) {
		int idx = 0;
		int nbsel = 0;
		// Count examples for fold
		for (int i = 0; i < m_NbRows; i++)
			if (m_Data[i].m_Folds[fold] != 0) nbsel++;
		// Select examples
		RowData res = new RowData(m_Schema, nbsel);
		for (int i = 0; i < m_NbRows; i++) {
			DataTuple tuple = m_Data[i];
			int factor = m_Data[i].m_Folds[fold];
			if (factor != 0) {
				DataTuple t2 = factor == 1 ? tuple : tuple.changeWeight(tuple.getWeight() * factor);
				res.setTuple(t2, idx++);
			}
		}
		return res;
	}

	public final RowData getOVFoldData(int fold) {
		int idx = 0;
		int nbsel = 0;
		// Count examples for fold
		for (int i = 0; i < m_NbRows; i++) {
			int efold = m_Data[i].m_Index;
			if (efold != -1) {
				if (efold != fold) nbsel++;
			} else {
				if (Arrays.binarySearch(m_Data[i].m_Folds, fold) >= 0) nbsel++;
			}
		}
		// Select examples
		RowData res = new RowData(m_Schema, nbsel);
		for (int i = 0; i < m_NbRows; i++) {
			DataTuple tuple = m_Data[i];
			int efold = tuple.m_Index;
			if (efold != -1) {
				if (efold != fold) res.setTuple(tuple, idx++);
			} else {
				if (Arrays.binarySearch(m_Data[i].m_Folds, fold) >= 0) res.setTuple(tuple, idx++);
			}
		}
		return res;
	}

	public final boolean checkData() {
		for (int i = 0; i < m_NbRows; i++) {
			DataTuple tuple = m_Data[i];
			if (tuple.m_Index == -1 && tuple.m_Folds == null) return false;
			if (tuple.m_Index != -1 && tuple.m_Folds != null) return false;
		}
		return true;
	}

	public final DataTuple getTuple(int i) {
		return m_Data[i];
	}

	public final void setTuple(DataTuple tuple, int i) {
		m_Data[i] = tuple;
	}



	public final RowData applyWeighted(NodeTest test, int branch) {
		int nb = 0;
		for (int i = 0; i < m_NbRows; i++) {
			int pred = test.predictWeighted(m_Data[i]);
			if (pred == branch || pred == NodeTest.UNKNOWN) nb++;
		}
		int idx = 0;
		RowData res = new RowData(m_Schema, nb);
		float prop = test.getProportion(branch);
		for (int i = 0; i < m_NbRows; i++) {
			DataTuple tuple = m_Data[i];
			int pred = test.predictWeighted(tuple);
			if (pred == branch) {
				res.setTuple(tuple, idx++);
			} else if (pred == NodeTest.UNKNOWN) {
				DataTuple ntuple = tuple.multiplyWeight(prop);
				res.setTuple(ntuple, idx++);
			}
		}
		return res;
	}

	public final RowData apply(NodeTest test, int branch) {
		int nb = 0;
		for (int i = 0; i < m_NbRows; i++) {
			int pred = test.predictWeighted(m_Data[i]);
			if (pred == branch) nb++;
		}
		int idx = 0;
		RowData res = new RowData(m_Schema, nb);
		for (int i = 0; i < m_NbRows; i++) {
			DataTuple tuple = m_Data[i];
			int pred = test.predictWeighted(tuple);
			if (pred == branch) res.setTuple(tuple, idx++);
		}
		return res;
	}

	public final RowData applySoft(SoftTest test, int branch) {
		int nb = 0;
		for (int i = 0; i < m_NbRows; i++)
			nb += test.softPredictNb(m_Data[i], branch);
		int idx = 0;
		RowData res = new RowData(m_Schema, nb);
		for (int i = 0; i < m_NbRows; i++)
			idx = test.softPredict(res, m_Data[i], idx, branch);
		return res;
	}

	public final RowData applySoft2(SoftTest test, int branch) {
		int nb = 0;
		for (int i = 0; i < m_NbRows; i++)
			nb += test.softPredictNb2(m_Data[i], branch);
		int idx = 0;
		RowData res = new RowData(m_Schema, nb);
		for (int i = 0; i < m_NbRows; i++)
			idx = test.softPredict2(res, m_Data[i], idx, branch);
		return res;
	}

	public void resize(int nbrows) {
		m_Data = new DataTuple[nbrows];
		for (int i = 0; i < nbrows; i++) m_Data[i] = new DataTuple(m_Schema);
		m_NbRows = nbrows;
	}

	public void resizeEmpty(int nbrows) {
		m_Data = new DataTuple[nbrows];
		m_NbRows = nbrows;
	}

	public void showDebug(ClusSchema schema) {
		System.out.println("Data: "+m_NbRows+" Size: "+m_Data.length);
		for (int i = 0; i < m_NbRows; i++) {
			DataTuple tuple = getTuple(i);
			if (tuple == null) {
				System.out.println("? ");
			} else {
				ClusAttrType at = schema.getAttrType(0);
				System.out.println(at.getString(tuple));
/*				if (tuple.m_Index == -1) {
					System.out.println(" Folds: "+MyIntArray.print(tuple.m_Folds));
				} else {
					System.out.println(" LO: "+tuple.m_Index);
				}*/
			}
		}
		System.out.println();
	}

	public void attach(ClusNode node) {
	}

	public void calcTotalStatBitVector(ClusStatistic stat) {
		stat.setSDataSize(getNbRows());
		calcTotalStat(stat);
		stat.optimizePreCalc(this);
	}

	public void calcTotalStat(ClusStatistic stat) {
		for (int i = 0; i < m_NbRows; i++) {
			stat.updateWeighted(m_Data[i], i);
		}
	}

	public final void calcPosAndMissStat(NodeTest test, int branch, ClusStatistic pos, ClusStatistic miss) {
		for (int i = 0; i < m_NbRows; i++) {
			DataTuple tuple = m_Data[i];
			int pred = test.predictWeighted(tuple);
			if (pred == branch) {
				pos.updateWeighted(m_Data[i], i);
			} else if (pred == NodeTest.UNKNOWN) {
				miss.updateWeighted(m_Data[i], i);
			}
		}
	}

	public final boolean isSoft() {
		for (int i = 0; i < m_NbRows; i++) {
			DataTuple tuple = m_Data[i];
			if (tuple.m_Index == -1) return true;
		}
		return false;
	}

	public final void calcXValTotalStat(ClusStatistic[] tot) {
		for (int i = 0; i < m_NbRows; i++) {
			DataTuple tuple = m_Data[i];
			tot[tuple.getIndex()].updateWeighted(tuple, i);
		}
	}

	public final void calcXValTotalStat(ClusStatistic[] tot, ClusStatistic[] extra) {
		for (int i = 0; i < extra.length; i++) extra[i].reset();
		for (int i = 0; i < m_NbRows; i++) {
			DataTuple tuple = m_Data[i];
			if (tuple.m_Index != -1) {
				tot[tuple.m_Index].updateWeighted(tuple, i);
			} else {
				int[] folds = tuple.m_Folds;
				for (int j = 0; j < folds.length; j++)
					extra[folds[j]].updateWeighted(tuple, i);
			}
		}
	}

	public void calcError(ClusNode node, ClusErrorList par) {
		for (int i = 0; i < m_NbRows; i++) {
			DataTuple tuple = getTuple(i);
			ClusStatistic stat = node.predictWeighted(tuple);
			par.addExample(tuple, stat);
		}
	}

	public void preprocess(int pass, DataPreprocs pps) throws ClusException {
		for (int i = 0; i < m_NbRows; i++) {
			DataTuple tuple = m_Data[i];
			pps.preproc(pass, tuple);
		}
	}

	public final void showTable() {
		for (int i = 0; i < m_Schema.getNbAttributes(); i++) {
			ClusAttrType type = m_Schema.getAttrType(i);
			if (i != 0) System.out.print(",");
			System.out.print(type.getName());
		}
		System.out.println();
		for (int i = 0; i < m_NbRows; i++) {
			DataTuple tuple = getTuple(i);
			for (int j = 0; j < m_Schema.getNbAttributes(); j++) {
				ClusAttrType type = m_Schema.getAttrType(j);
				if (j != 0) System.out.print(",");
				System.out.print(type.getString(tuple));
			}
			System.out.println();
		}
	}

	public float[] getNumeric(int idx) {
		return m_Data[idx].m_floats;
	}

	public int[] getNominal(int idx) {
		return m_Data[idx].m_Ints;
	}

	public MemoryTupleIterator getIterator() {
		return new MemoryTupleIterator(this);
	}

	public void addIndices() {
		for (int i = 0; i < m_NbRows; i++) {
			m_Data[i].setIndex(i);
		}
	}

	public boolean equals(RowData d){
		if (d.getNbRows() != getNbRows()) {
			return false;
		}
		if (!m_Schema.equals(m_Schema)) {
			return false;
		}
		for (int i = 0; i < getNbRows(); i++) {
			if (!d.getTuple(i).equals(getTuple(i))) return false;
		}
		return false;
	}

	public RowData sampleWeighted(Random random) {
		return sampleWeighted(random, getNbRows());
	}
	
	public RowData sampleWeighted(Random random, int N) {		
		float[] weight_acc = new float[getNbRows()];
		weight_acc[0] = getTuple(0).getWeight();
		for (int i = 1; i < getNbRows(); i++) {
			DataTuple tuple = getTuple(i);
			weight_acc[i] = weight_acc[i-1] + tuple.getWeight();
		}
		float tot_w = weight_acc[getNbRows()-1];
		ArrayList res = new ArrayList();
		for (int i = 0; i < N; i++) {
			float rnd = random.nextFloat()*tot_w;		
			// Index of the search key, if it is contained in the list; otherwise, (-(insertion point) - 1). 
			// The insertion point is defined as the point at which the key would be inserted into the list: 
			// the index of the first element greater than the key, or list.size(), if all elements in the list
			// are less than the specified key		
			int loc = Arrays.binarySearch(weight_acc, rnd);
			if (loc < 0) {
				loc = -loc - 1;
			}
			DataTuple restuple = getTuple(loc).changeWeight(1.0f);
			res.add(restuple);
		}
		return new RowData(res, getSchema());
	}	
	
	public class MySortableArray implements MSortable {

		DoubleObject[] tuples;

		public MySortableArray(DoubleObject[] data) {
			tuples = data;
		}

		public float getfloat(int i) {
			return tuples[i].getValue();
		}

		public void swap(int i, int j) {
			DoubleObject obj_i = tuples[i];
			tuples[i] = tuples[j];
			tuples[j] = obj_i;
		}
	}
}
