/*************************************************************************
 * Clus - Software for Predictive Clustering                             *
 * Copyright (C) 2007                                                    *
 *    Katholieke Universiteit Leuven, Leuven, Belgium                    *
 *    Jozef Stefan Institute, Ljubljana, Slovenia                        *
 *                                                                       *
 * This program is free software: you can redistribute it and/or modify  *
 * it under the terms of the GNU General Public License as published by  *
 * the Free Software Foundation, either version 3 of the License, or     *
 * (at your option) any later version.                                   *
 *                                                                       *
 * This program is distributed in the hope that it will be useful,       *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 * GNU General Public License for more details.                          *
 *                                                                       *
 * You should have received a copy of the GNU General Public License     *
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 *                                                                       *
 * Contact information: <http://www.cs.kuleuven.be/~dtai/clus/>.         *
 *************************************************************************/

/*
 * Created on May 12, 2005
 */
package clus.error;

import java.io.PrintWriter;
import java.text.NumberFormat;
import clus.data.attweights.*;
import clus.data.rows.*;
import clus.data.type.*;
import clus.main.*;
import clus.statistic.*;

// import jeans.util.array.*;

public class MSError extends ClusNumericError {

	public final static long serialVersionUID = Settings.SERIAL_VERSION_ID;
	
	
	protected int[]	   m_nbEx;

	protected float[] m_SumErr;
	protected float[] m_SumSqErr;
	protected ClusAttributeWeights m_Weights;
	protected boolean m_PrintAllComps;

	public MSError(ClusErrorList par, NumericAttrType[] num) {
		this(par, num, null, true);
	}

	public MSError(ClusErrorList par, NumericAttrType[] num, ClusAttributeWeights weights) {
		this(par, num, weights, true);
	}

	public MSError(ClusErrorList par, NumericAttrType[] num, ClusAttributeWeights weights, boolean printall) {
		super(par, num);
		m_nbEx = new int[m_Dim];
		m_SumErr = new float[m_Dim];
		m_SumSqErr = new float[m_Dim];
		m_Weights = weights;
		m_PrintAllComps = printall;
	}

	public void reset() {
		for (int i = 0; i < m_Dim; i++) {
			m_SumErr[i] = 0.0f;
			m_SumSqErr[i] = 0.0f;
			m_nbEx[i] = 0;
		}
	}

	public void setWeights(ClusAttributeWeights weights) {
		m_Weights = weights;
	}

	public float getModelErrorComponent(int i) {
		
		//int nb = getNbExamples();
		int nb = m_nbEx[i];

		//System.out.println(m_SumErr[i]);
		float err = nb != 0.0 ? m_SumErr[i]/nb : 0.0f;
		if (m_Weights != null) err *= m_Weights.getWeight(getAttr(i));

		return err;
	}

	public float getModelError() {
		float ss_tree = 0.0f;
		int nb = 0;
		for(int i=0;i<m_Dim;i++){
			nb+= m_nbEx[i];
		}
		if (m_Weights != null) {
			for (int i = 0; i < m_Dim; i++) {
				ss_tree += m_SumErr[i]*m_Weights.getWeight(getAttr(i));
			}
			return nb != 0.0 ? ss_tree/nb : 0.0f;
		} else {
			for (int i = 0; i < m_Dim; i++) {
				ss_tree += m_SumErr[i];
			}
			return nb != 0.0 ? ss_tree/nb : 0.0f;
		}
	}

	public float getModelErrorStandardError() {
		float sum_err = 0.0f;
		float sum_sq_err = 0.0f;
		for (int i = 0; i < m_Dim; i++) {
			if (m_Weights != null) {
				sum_err += m_SumErr[i];
				sum_sq_err += m_SumSqErr[i];
			} else {
				sum_err += m_SumErr[i]*m_Weights.getWeight(getAttr(i));
				sum_sq_err += m_SumSqErr[i]*sqr(m_Weights.getWeight(getAttr(i)));
			}
		}
		//float n = getNbExamples() * m_Dim;
		float n = 0;
		for(int i=0;i<m_Dim;i++){
			n+= m_nbEx[i];
		}
		
		if (n <= 1) {
			return Float.POSITIVE_INFINITY;
		} else {
			float ss_x = (n * sum_sq_err - sqr(sum_err)) / (n * (n-1));
			return(float) Math.sqrt(ss_x / n);
		}
	}

	public final static float sqr(float value) {
		return value*value;
	}

	public void addExample(float[] real, float[] predicted) {
		for (int i = 0; i < m_Dim; i++) {
			float err = sqr(real[i] - predicted[i]);
			System.out.println(err);
			if(!Float.isInfinite(err) && !Float.isNaN(err)){
				m_SumErr[i] += err;
				m_SumSqErr[i] += sqr(err);
				m_nbEx[i]++;
			}
		}
	}

	public void addExample(DataTuple tuple, ClusStatistic pred) {
		float[] predicted = pred.getNumericPred();
		for (int i = 0; i < m_Dim; i++) {
			float err = sqr(getAttr(i).getNumeric(tuple) - predicted[i]);			
			if(!Float.isInfinite(err) && !Float.isNaN(err)){
				m_SumErr[i] += err;
				m_SumSqErr[i] += sqr(err);
				m_nbEx[i]++;
			}
		}		
	}

	public void addExample(DataTuple real, DataTuple pred) {
		for (int i = 0; i < m_Dim; i++) {
				float real_i = getAttr(i).getNumeric(real);
				float predicted_i = getAttr(i).getNumeric(pred);
				float err = sqr(real_i - predicted_i);
				if(!Float.isInfinite(err) && !Float.isNaN(err)){
					m_SumErr[i] += err;
					m_SumSqErr[i] += sqr(err);
					m_nbEx[i]++;
				}
		}		
	}	
	public void addInvalid(DataTuple tuple) {
	}

	public void add(ClusError other) {
		MSError oe = (MSError)other;
		for (int i = 0; i < m_Dim; i++) {
			m_SumErr[i] += oe.m_SumErr[i];
			m_SumSqErr[i] += oe.m_SumSqErr[i];
			m_nbEx[i]+= oe.m_nbEx[i];
		}
	}

	public void showModelError(PrintWriter out, int detail) {
		NumberFormat fr = getFormat();
		StringBuffer buf = new StringBuffer();
		if (m_PrintAllComps) {
			buf.append("[");
			for (int i = 0; i < m_Dim; i++) {
				if (i != 0) buf.append(",");
				buf.append(fr.format(getModelErrorComponent(i)));
			}
			if (m_Dim > 1) buf.append("]: ");
			else buf.append("]");
		}
		if (m_Dim > 1 || !m_PrintAllComps) {
			buf.append(fr.format(getModelError()));
		}
		out.println(buf.toString());
	}

	public void showSummaryError(PrintWriter out, boolean detail) {
		NumberFormat fr = getFormat();
		out.println(getPrefix() + "Mean over components MSE: "+fr.format(getModelError()));
	}

	public String getName() {
		if (m_Weights == null) return "Mean squared error (MSE)";
		else return "Weighted mean squared error (MSE) ("+m_Weights.getName(m_Attrs)+")";
	}

	public ClusError getErrorClone(ClusErrorList par) {
		return new MSError(par, m_Attrs, m_Weights, m_PrintAllComps);
	}

	public float computeLeafError(ClusStatistic stat) {
		RegressionStat rstat = (RegressionStat)stat;
		return rstat.getSS(m_Weights) * rstat.getNbAttributes();
	}
}
