package clus.error;

import java.io.PrintWriter;
import java.text.NumberFormat;

import clus.data.attweights.ClusAttributeWeights;
import clus.data.rows.DataTuple;
import clus.data.type.NominalAttrType;
import clus.statistic.ClassificationStat;
import clus.statistic.ClusStatistic;

public class MSNominalError extends ClusNominalError {
	
	public final static long serialVersionUID = 1L;

	protected ClusAttributeWeights m_Weights;
	protected float[] m_SumErr;
	protected float[] m_SumSqErr;
	protected boolean m_PrintAllComps = true;

	public MSNominalError(ClusErrorList par, NominalAttrType[] nom, ClusAttributeWeights weights) {
		super(par, nom);
		m_Weights = weights;
		m_SumErr = new float[m_Dim];
		m_SumSqErr = new float[m_Dim];
	}

	public ClusError getErrorClone(ClusErrorList par) {
		return new MSNominalError(par, m_Attrs, m_Weights);
	}

	public String getName() {
		if (m_Weights == null) return "Mean squared error (MSE) for Nominal Attributes";
		else return "Weighted mean squared error (MSE) for Nominal Attributes ("+m_Weights.getName(m_Attrs)+")";
	}

	public float getModelErrorComponent(int i) {
		int nb = getNbExamples();
		float err = nb != 0.0 ? m_SumErr[i]/nb : 0.0f;
		if (m_Weights != null) err *= m_Weights.getWeight(getAttr(i));
		return err;
	}

	public float getModelError() {
		float ss_tree = 0.0f;
		int nb = getNbExamples();
		if (m_Weights != null) {
			for (int i = 0; i < m_Dim; i++) {
				ss_tree += m_SumErr[i]*m_Weights.getWeight(getAttr(i));
			}
			return nb != 0.0 ? ss_tree/nb/m_Dim : 0.0f;
		} else{
			for (int i = 0; i < m_Dim; i++) {
				ss_tree += m_SumErr[i];
			}
			return nb != 0.0 ? ss_tree/nb/m_Dim : 0.0f;
		}
	}

	public float getModelErrorStandardError() {
		float sum_err = 0.0f;
		float sum_sq_err = 0.0f;
		for (int i = 0; i < m_Dim; i++) {
			if (m_Weights != null) {
				sum_err += m_SumErr[i];
				sum_sq_err += m_SumSqErr[i];
			} else {
				sum_err += m_SumErr[i]*m_Weights.getWeight(getAttr(i));
				sum_sq_err += m_SumSqErr[i]*sqr(m_Weights.getWeight(getAttr(i)));
			}
		}
		float n = getNbExamples() * m_Dim;
		if (n <= 1) {
			return Float.POSITIVE_INFINITY;
		} else {
			float ss_x = (n * sum_sq_err - sqr(sum_err)) / (n * (n-1));
			return(float) Math.sqrt(ss_x / n);
		}
	}

	public final static float sqr(float value) {
		return value*value;
	}

	public void addExample(DataTuple tuple, ClusStatistic pred) {
		ClassificationStat stat = pred.getClassificationStat();
		for (int i = 0; i < m_Dim; i++) {
			NominalAttrType type = m_Attrs[i];
			int value = type.getNominal(tuple);
			for (int j = 0; j < type.getNbValues(); j++) {
				float zeroOne = (value == j) ? 1.0f : 0.0f;
				float err = sqr(zeroOne - stat.getCount(i, j)/stat.getTotalWeight());
				m_SumErr[i] += err;
				m_SumSqErr[i] += sqr(err);
			}
		}
	}

	public void addInvalid(DataTuple tuple) {
	}

	public void reset() {
		for (int i = 0; i < m_Dim; i++) {
			m_SumErr[i] = 0.0f;
			m_SumSqErr[i] = 0.0f;
		}
	}

	public void add(ClusError other) {
		MSNominalError oe = (MSNominalError)other;
		for (int i = 0; i < m_Dim; i++) {
			m_SumErr[i] += oe.m_SumErr[i];
			m_SumSqErr[i] += oe.m_SumSqErr[i];
		}
	}

	public void showModelError(PrintWriter out, int detail) {
		NumberFormat fr = getFormat();
		StringBuffer buf = new StringBuffer();
		if (m_PrintAllComps) {
			buf.append("[");
			for (int i = 0; i < m_Dim; i++) {
				if (i != 0) buf.append(",");
				buf.append(fr.format(getModelErrorComponent(i)));
			}
			if (m_Dim > 1) buf.append("]: ");
			else buf.append("]");
		}
		if (m_Dim > 1 || !m_PrintAllComps) {
			buf.append(fr.format(getModelError()));
		}
		out.println(buf.toString());
	}

	public void showSummaryError(PrintWriter out, boolean detail) {
		NumberFormat fr = getFormat();
		out.println(getPrefix() + "Mean over components MSE: "+fr.format(getModelError()));
	}

	public float computeLeafError(ClusStatistic stat) {
		ClassificationStat cstat = (ClassificationStat)stat;
		return cstat.getSS(m_Weights) * cstat.getNbAttributes();
	}
}
