/*************************************************************************
 * Clus - Software for Predictive Clustering                             *
 * Copyright (C) 2007                                                    *
 *    Katholieke Universiteit Leuven, Leuven, Belgium                    *
 *    Jozef Stefan Institute, Ljubljana, Slovenia                        *
 *                                                                       *
 * This program is free software: you can redistribute it and/or modify  *
 * it under the terms of the GNU General Public License as published by  *
 * the Free Software Foundation, either version 3 of the License, or     *
 * (at your option) any later version.                                   *
 *                                                                       *
 * This program is distributed in the hope that it will be useful,       *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 * GNU General Public License for more details.                          *
 *                                                                       *
 * You should have received a copy of the GNU General Public License     *
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 *                                                                       *
 * Contact information: <http://www.cs.kuleuven.be/~dtai/clus/>.         *
 *************************************************************************/

package clus.heuristic;

import clus.main.*;
import clus.util.*;
import jeans.math.*;

import java.util.*;

import org.apache.commons.math.*;
import org.apache.commons.math.distribution.*;

public class FTest {

	public static float[] FTEST_SIG = {1.0f, 0.1f, 0.05f, 0.01f, 0.005f, 0.001f, 0.0f};
	public static float FTEST_LIMIT;
	public static float[] FTEST_VALUE;

	public final static FDistribution m_FDist = DistributionFactory.newInstance().createFDistribution(1,1);

	protected final static float critical_f_01[] = {
		39.8161f, 8.5264f, 5.5225f, 4.5369f, 4.0804f, 3.7636f,
		3.61f, 3.4596f, 3.3489f, 3.2761f, 3.24f, 3.1684f, 3.1329f, 3.0976f,
		3.0625f, 3.0625f, 3.0276f, 2.9929f, 2.9929f, 2.9584f
	};

	protected final static float critical_f_005[] = {
		161.0f, 18.5f, 10.1f, 7.71f, 6.61f, 5.99f, 5.59f, 5.32f, 5.12f, 4.96f,
		4.84f, 4.75f, 4.67f, 4.6f, 4.54f, 4.49f, 4.45f, 4.41f, 4.38f, 4.35f,
		4.32f, 4.3f, 4.28f, 4.26f, 4.24f, 4.23f, 4.21f, 4.2f, 4.18f, 4.17f
	};

	protected final static float critical_f_001[] = {
		4052.0f, 98.5f, 34.1f, 21.2f, 16.3f, 13.7f, 12.2f, 11.3f, 10.6f, 10.0f,
		9.65f, 9.33f, 9.07f, 8.86f, 8.68f, 8.53f, 8.40f, 8.29f, 8.18f, 8.1f,
		8.02f, 7.95f, 7.88f, 7.82f, 7.77f, 7.72f, 7.68f, 7.64f, 7.6f, 7.56f
	};

	// for 0.0f05: derived from t-table,  squared values (=> approximate!)
	protected final static float critical_f_0005[] = {
		15876f, 198.81f, 55.5025f, 31.36f, 22.7529f, 18.6624f, 16.2409f, 14.6689f, 13.6161f, 12.8164f,
		12.25f, 11.7649f, 11.2896f, 11.0889f, 10.8241f, 10.5625f, 10.3684f, 10.24f, 10.0489f, 9.9225f,
		9.8596f, 9.7344f, 9.61f, 9.5481f, 9.4864f, 9.4249f, 9.3636f, 9.3025f, 9.2416f, 9.1809f
	};

	// for 0.0f01: derived from t-table,  squared values (=> approximate!)
	protected final static float critical_f_0001[] = {
		405769f, 998.56f, 166.41f, 74.1321f, 47.0596f, 35.5216f, 29.16f, 25.4016f, 22.8484f, 21.0681f,
		19.7136f, 18.6624f, 17.8084f, 17.1396f, 16.5649f, 16.0801f, 15.6025f, 15.3664f, 15.0544f, 14.8225f,
		14.5924f, 14.3641f, 14.2129f, 13.9876f, 13.8384f, 13.7641f, 13.6161f, 13.4689f, 13.3956f, 13.3225f
	};

	public static int getLevelAndComputeArray(float significance) {
		int maxlevel = FTEST_SIG.length-1;
		for (int level = 0; level < maxlevel; level++) {
			if (Math.abs(significance - FTEST_SIG[level])/FTEST_SIG[level] < 0.01f) {
				return level;
			}
		}
		FTEST_SIG[maxlevel] = significance;
		initializeFTable(significance);
		return maxlevel;
	}

	public static float getCriticalF(int level, int df) {
		switch (level) {
			case 1: // 0.1
				if (df <= 20) return critical_f_01[df-1];
				else if (df <= 30) return 2.9f;
				else if (df <= 40) return 2.86f;
				else if (df <= 120) return 2.79f;
				else return 2.7f;
			case 2:	// 0.0f5
				if (df <= 30) return critical_f_005[df-1];
				else if (df <= 40) return 4.08f;
				else if (df <= 60) return 4.00f;
				else if (df <= 120) return 3.92f;
				else return 3.84f;
			case 3:	// 0.0f1
				if (df <= 30) return critical_f_001[df-1];
				else if (df <= 40) return 7.31f;
				else if (df <= 60) return 7.08f;
				else if (df <= 120) return 6.85f;
				else return 6.63f;
			case 4: // 0.0f05
				if (df <= 30) return critical_f_0005[df-1];
				else if (df <= 40) return 8.82f;
				else if (df <= 60) return 8.47f;
				else if (df <= 120) return 8.18f;
				else return 7.90f;
			case 5: // 0.0f01
				if (df <= 30) return critical_f_0001[df-1];
				else if (df <= 40) return 12.60f;
				else if (df <= 60) return 11.98f;
				else if (df <= 120) return 11.36f;
				else return 10.82f;
			default:
				return df < FTEST_VALUE.length ? FTEST_VALUE[df] : FTEST_LIMIT;
		}
	}

	public static float getCriticalFCommonsMath(float sig, float df) {
		try {
			m_FDist.setDenominatorDegreesOfFreedom(df);
			return (float)m_FDist.inverseCumulativeProbability(1-sig);
		} catch (MathException e) {
			System.err.println("F-Distribution error: "+e.getMessage());
			return 0.0f;
		}
	}

	// Calling getCriticalFCommonsMath() is slow, so build a table
	public static void initializeFTable(float sig) {
		int df = 3;
		float value = 0.0f;
		float limit = (float)getCriticalFCommonsMath(sig, 1e5f);
		ArrayList values = new ArrayList();
		do {
			value = getCriticalFCommonsMath(sig, df);
			values.add(new Float(value));
			df++;
		} while ((value - limit)/limit > 0.05f);
		System.out.println("F-Test = "+sig+" limit = "+ClusFormat.TWO_AFTER_DOT.format(limit)+" values = "+values.size());
		FTEST_LIMIT = limit;
		FTEST_VALUE = new float[values.size()+3];
		for (int i = 0; i < values.size(); i++) {
			FTEST_VALUE[i+3] = ((Float)values.get(i)).floatValue();
		}
	}

	// ftest: Signif, total SS, residual SS, 2nd DF (1st is 1)
	// (this implementation only works for F tests with 1 and n d.f., sorry)
	// ftest predicate succeeds iff H0 is rejected at Signif
	// only works correctly for signif 1.0f, 0.1, 0.0f5, 0.0f1, 0.0f05, 0.0f01

	public static boolean ftest(int level, float sst, float ssr, int df) {
		if (level == 0) return true;
		if (sst <= 0.0f) return false; // added to avoid 0/0; SST = 0 => no improvement possible
		if (ssr == 0.0f) return true;  // avoid x/0; SSR = 0 => F is infinite
		float f = (float)df * (sst - ssr) / ssr;
		float cf = getCriticalF(level, df);
		return f > cf;
	}

	public static float getSettingSig() {
		return FTEST_SIG[Settings.FTEST_LEVEL];
	}

	public static float transformHeur(float value, float n_tot, float n_pos, float n_neg) {
		if (value < 1e-6) return Float.NEGATIVE_INFINITY;
		else return value;
	}

	public static float calcSSHeuristicRatio(float ss_tot, float sum_ss, float n_tot, float n_pos, float n_neg) {
		float val = ss_tot - sum_ss;
		int n_2 = (int)Math.floor(n_tot - 2.0f + 0.5f);
		if (n_2 <= 0) {
			if (Settings.FTEST_LEVEL == 0) return transformHeur(val, n_tot, n_pos, n_neg);
			else return Float.NEGATIVE_INFINITY;
		} else {
			if (ftest(Settings.FTEST_LEVEL, ss_tot, sum_ss, n_2)) {
				return transformHeur(val, n_tot, n_pos, n_neg);
			} else {
				return Float.NEGATIVE_INFINITY;
			}
		}
	}

	public static float calcSSHeuristic(float n_tot, float ss_tot, float ss_sum) {
		int n_2 = (int)Math.floor(n_tot - 2.0f + 0.5f);
		float value = ss_tot - ss_sum;
		if (value < MathUtil.C1E_9) return Float.NEGATIVE_INFINITY;
		if (n_2 <= 0) {
			if (Settings.FTEST_LEVEL == 0) return value;
			else return Float.NEGATIVE_INFINITY;
		} else {
			if (FTest.ftest(Settings.FTEST_LEVEL, ss_tot, ss_sum, n_2)) {
				return value;
			} else {
				return Float.NEGATIVE_INFINITY;
			}
		}
	}	
}
