/*************************************************************************
 * Clus - Software for Predictive Clustering                             *
 * Copyright (C) 2007                                                    *
 *    Katholieke Universiteit Leuven, Leuven, Belgium                    *
 *    Jozef Stefan Institute, Ljubljana, Slovenia                        *
 *                                                                       *
 * This program is free software: you can redistribute it and/or modify  *
 * it under the terms of the GNU General Public License as published by  *
 * the Free Software Foundation, either version 3 of the License, or     *
 * (at your option) any later version.                                   *
 *                                                                       *
 * This program is distributed in the hope that it will be useful,       *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 * GNU General Public License for more details.                          *
 *                                                                       *
 * You should have received a copy of the GNU General Public License     *
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 *                                                                       *
 * Contact information: <http://www.cs.kuleuven.be/~dtai/clus/>.         *
 *************************************************************************/

package clus.statistic;

import java.io.IOException;
import java.io.PrintWriter;
import java.text.*;
import java.util.ArrayList;

import jeans.util.StringUtils;

import org.apache.commons.math.MathException;
import org.apache.commons.math.distribution.*;

import clus.main.ClusStatManager;
import clus.main.Settings;
import clus.util.*;
import clus.data.cols.*;
import clus.data.rows.*;
import clus.data.type.*;
import clus.data.attweights.*;
import clus.error.ClusNumericError;

public class RegressionStat extends ClusStatistic {

	public final static long serialVersionUID = Settings.SERIAL_VERSION_ID;

	public int m_NbAttrs;
	public float[] m_SumValues;
	public float[] m_SumWeights;
	public float[] m_SumSqValues;
	public float[] m_Means;
	public NumericAttrType[] m_Attrs;

	public RegressionStat(NumericAttrType[] attrs) {
		this(attrs, false);
	}

	public RegressionStat(NumericAttrType[] attrs, boolean onlymean) {
		m_Attrs = attrs;
		m_NbAttrs = attrs.length;
		if (onlymean) {
			m_Means = new float[m_NbAttrs];
		} else {
			m_SumValues = new float[m_NbAttrs];
			m_SumWeights = new float[m_NbAttrs];
			m_SumSqValues = new float[m_NbAttrs];
		}
	}

	public ClusStatistic cloneStat() {
		return new RegressionStat(m_Attrs, false);
	}

	public ClusStatistic cloneSimple() {
		return new RegressionStat(m_Attrs, true);
	}

	public int getNbAttributes() {
		return m_NbAttrs;
	}

	public NumericAttrType getAttribute(int idx) {
		return m_Attrs[idx];
	}

	public void reset() {
		m_SumWeight = 0.0f;
		m_nbEx = 0;
		for (int i = 0; i < m_NbAttrs; i++) {
			m_SumWeights[i] = 0.0f;
			m_SumValues[i] = 0.0f;
			m_SumSqValues[i] = 0.0f;
		}
	}

	public void copy(ClusStatistic other) {
		RegressionStat or = (RegressionStat)other;
		m_SumWeight = or.m_SumWeight;
		m_nbEx = or.m_nbEx;
		System.arraycopy(or.m_SumWeights, 0, m_SumWeights, 0, m_NbAttrs);
		System.arraycopy(or.m_SumValues, 0, m_SumValues, 0, m_NbAttrs);
		System.arraycopy(or.m_SumSqValues, 0, m_SumSqValues, 0, m_NbAttrs);
	}
	
	/**
	 * Used for combining weighted predictions. 
	 */
	public RegressionStat normalizedCopy() {
		RegressionStat copy = (RegressionStat)cloneSimple();
		copy.m_nbEx = 0;
		copy.m_SumWeight = 1;
		calcMean(copy.m_Means);
		return copy;
	}	

	public void addPrediction(ClusStatistic other, float weight) {
		RegressionStat or = (RegressionStat)other;
		for (int i = 0; i < m_NbAttrs; i++) {
			m_Means[i] += weight*or.m_Means[i];
		}
	}

	public void add(ClusStatistic other) {
		RegressionStat or = (RegressionStat)other;
		m_SumWeight += or.m_SumWeight;
		m_nbEx += or.m_nbEx;
		for (int i = 0; i < m_NbAttrs; i++) {
			m_SumWeights[i] += or.m_SumWeights[i];
			m_SumValues[i] += or.m_SumValues[i];
			m_SumSqValues[i] += or.m_SumSqValues[i];
		}
	}

	public void addScaled(float scale, ClusStatistic other) {
		RegressionStat or = (RegressionStat)other;
		m_SumWeight += scale * or.m_SumWeight;
		m_nbEx += or.m_nbEx;
		for (int i = 0; i < m_NbAttrs; i++) {
			m_SumWeights[i] += scale * or.m_SumWeights[i];
			m_SumValues[i] += scale * or.m_SumValues[i];
			m_SumSqValues[i] += scale * or.m_SumSqValues[i];
		}
	}

	public void subtractFromThis(ClusStatistic other) {
		RegressionStat or = (RegressionStat)other;
		m_SumWeight -= or.m_SumWeight;
		m_nbEx -= or.m_nbEx;
		for (int i = 0; i < m_NbAttrs; i++) {
			m_SumWeights[i] -= or.m_SumWeights[i];
			m_SumValues[i] -= or.m_SumValues[i];
			m_SumSqValues[i] -= or.m_SumSqValues[i];
		}
	}

	public void subtractFromOther(ClusStatistic other) {
		RegressionStat or = (RegressionStat)other;
		m_SumWeight = or.m_SumWeight - m_SumWeight;
		m_nbEx = or.m_nbEx - m_nbEx;
		for (int i = 0; i < m_NbAttrs; i++) {
			m_SumWeights[i] = or.m_SumWeights[i] - m_SumWeights[i];
			m_SumValues[i] = or.m_SumValues[i] - m_SumValues[i];
			m_SumSqValues[i] = or.m_SumSqValues[i] - m_SumSqValues[i];
		}
	}

	public void update(ColTarget target, int idx) {
		m_nbEx += 1;
		m_SumWeight += 1.0f;
		float[] values = target.m_Numeric[idx];
		for (int i = 0; i < m_NbAttrs; i++) {
			float val = values[i];
			if (val != Float.POSITIVE_INFINITY) {
				m_SumWeights[i] += 1.0f;
				m_SumValues[i] += val;
				m_SumSqValues[i] += val*val;
			}
		}
	}

	public void updateWeighted(DataTuple tuple, int idx) {
		updateWeighted(tuple, tuple.getWeight());
	}

	public void updateWeighted(DataTuple tuple, float weight) {
		m_nbEx += 1;
		m_SumWeight += weight;
		for (int i = 0; i < m_NbAttrs; i++) {
			float val = m_Attrs[i].getNumeric(tuple);
			if (val != Float.POSITIVE_INFINITY) {
				m_SumWeights[i] += weight;
				m_SumValues[i] += weight*val;
				m_SumSqValues[i] += weight*val*val;
			}
		}
	}

	public void computePrediction() {
		// do not need to call calcMean() here
	}

	public void calcMean(float[] means) {
		for (int i = 0; i < m_NbAttrs; i++) {
			means[i] = m_SumWeights[i] != 0.0f ? m_SumValues[i] / m_SumWeights[i] : 0.0f;
		}
	}	
	
	public void calcMean() {
		if (m_Means == null) m_Means = new float[m_NbAttrs];
		calcMean(m_Means);
	}
	
	public float getMean(int i) {
		return m_SumWeights[i] != 0.0f ? m_SumValues[i] / m_SumWeights[i] : 0.0f;
	}

	public float getSumValues(int i) {
		return m_SumValues[i];
	}
	
	public float getSumWeights(int i) {
		return m_SumWeights[i];
	}
		
	public float getSS(int i) {
		float n_tot = m_SumWeight;
		float k_tot = m_SumWeights[i];
		float sv_tot = m_SumValues[i];
		float ss_tot = m_SumSqValues[i];
		return (k_tot > 1.0f) ? ss_tot * (n_tot - 1) / (k_tot - 1) - n_tot * sv_tot/k_tot*sv_tot/k_tot : 0.0f;
	}

	public float getScaledSS(int i, ClusAttributeWeights scale) {
		return getSS(i)*scale.getWeight(getAttribute(i));
	}

	public float getVariance(int i) {
		return m_SumWeight != 0.0f ? getSS(i) / m_SumWeight : 0.0f;
	}

	public float getScaledVariance(int i, ClusAttributeWeights scale) {
		return getVariance(i)*scale.getWeight(getAttribute(i));
	}

	public float getRootScaledVariance(int i, ClusAttributeWeights scale) {
		return(float) Math.sqrt(getScaledVariance(i, scale));
	}

	public float[] getRootScaledVariances(ClusAttributeWeights scale) {
		int nb = getNbAttributes();
		float[] res = new float[nb];
		for (int i = 0; i < res.length; i++) {
			res[i] = getRootScaledVariance(i, scale);
		}
		return res;
	}

	public float getStandardDeviation(int i) {
		float n_tot = m_SumWeight;
		float k_tot = m_SumWeights[i];
		float sv_tot = m_SumValues[i];
		float ss_tot = m_SumSqValues[i];
		float var = (k_tot > 1.0f) ? ss_tot * (n_tot - 1) / (k_tot - 1) - n_tot * sv_tot/k_tot*sv_tot/k_tot : 0.0f;
		return(float) Math.sqrt(var / (n_tot - 1));
	}
	
	/**
	 * Currently only used to compute the default dispersion within rule heuristics.
	 */
	public float getDispersion(ClusAttributeWeights scale, RowData data) {
		System.err.println(getClass().getName()+": getDispersion(): Not yet implemented!");
		return Float.POSITIVE_INFINITY;
	}

	/**
	 * Computes a 2-sample t statistic, without the hypothesis of equal sub-population variances,
	 * and returns the p-value of a t test.
	 * t = (m1 - m2) / sqrt(var1/n1 + var2/n2);
	 * @param att attribute index
	 * @return t p-value
	 * @throws MathException
	 */
	public float getTTestPValue(int att, ClusStatManager stat_manager) throws MathException {
		float global_mean = ((CombStat)stat_manager.getTrainSetStat()).m_RegStat.getMean(att);
		float global_var = ((CombStat)stat_manager.getTrainSetStat()).m_RegStat.getVariance(att);
		float global_n = ((CombStat)stat_manager.getTrainSetStat()).getTotalWeight();
		float local_mean = getMean(att);
		float local_var = getVariance(att);
		float local_n = getTotalWeight();
		float t =(float) Math.abs(local_mean - global_mean) /(float) Math.sqrt(local_var/local_n + global_var/global_n);
		float degreesOfFreedom = 0;
		degreesOfFreedom = df(local_var, global_var, local_n, global_n);
		DistributionFactory distributionFactory = DistributionFactory.newInstance();
		TDistribution tDistribution = distributionFactory.createTDistribution(degreesOfFreedom);
		return 1.0f - (float)tDistribution.cumulativeProbability(-t, t);
	}

	/**
	 * Computes approximate degrees of freedom for 2-sample t-test.
	 * source:(float) Math.commons.stat.inference.TTestImpl
	 *
	 * @param v1 first sample variance
	 * @param v2 second sample variance
	 * @param n1 first sample n
	 * @param n2 second sample n
	 * @return approximate degrees of freedom
	 */
	protected float df(float v1, float v2, float n1, float n2) {
		return (float)(((v1 / n1) + (v2 / n2)) * ((v1 / n1) + (v2 / n2))) /
			((v1 * v1) / (n1 * n1 * (n1 - 1f)) + (v2 * v2) / (n2 * n2 * (n2 - 1f)));
	}

	/**
	 * @return Array for all the attributes.
	 */
	public float[] getNumericPred() {
		return m_Means;
	}

	public String getPredictedClassName(int idx) {
		return "";
	}

	public int getNbNumericAttributes() {
		return m_NbAttrs;
	}

	public float getError(ClusAttributeWeights scale) {
		return getSS(scale);
	}

	public float getErrorDiff(ClusAttributeWeights scale, ClusStatistic other) {
		return getSSDiff(scale, other);
	}

	public float getSS(ClusAttributeWeights scale) {
		float result = 0.0f;
		for (int i = 0; i < m_NbAttrs; i++) {
			float n_tot = m_SumWeight;
			float k_tot = m_SumWeights[i];
			float sv_tot = m_SumValues[i];
			float ss_tot = m_SumSqValues[i];
//			float wi = scale.getWeight(m_Attrs[i]);
			result += ((k_tot > 1.0f) ? ss_tot * (n_tot - 1) / (k_tot - 1) - n_tot * sv_tot/k_tot*sv_tot/k_tot : 0.0f)*scale.getWeight(m_Attrs[i]);
		}
		return result / m_NbAttrs;
	}

	public float getRMSE(ClusAttributeWeights scale) {
		return(float) Math.sqrt(getSS(scale)/getTotalWeight());
	}

	public float getSSDiff(ClusAttributeWeights scale, ClusStatistic other) {
		float result = 0.0f;
		RegressionStat or = (RegressionStat)other;
		for (int i = 0; i < m_NbAttrs; i++) {
			float n_tot = m_SumWeight - or.m_SumWeight;
			float k_tot = m_SumWeights[i] - or.m_SumWeights[i];
			float sv_tot = m_SumValues[i] - or.m_SumValues[i];
			float ss_tot = m_SumSqValues[i] - or.m_SumSqValues[i];
			result += ((k_tot > 1.0f) ? ss_tot * (n_tot - 1) / (k_tot - 1) - n_tot * sv_tot/k_tot*sv_tot/k_tot : 0.0f)*scale.getWeight(m_Attrs[i]);
		}
		return result / m_NbAttrs;
	}

	public void initNormalizationWeights(ClusAttributeWeights weights, boolean[] shouldNormalize) {
		for (int i = 0; i < m_NbAttrs; i++) {
			int idx = m_Attrs[i].getIndex();
			if (shouldNormalize[idx]) {
				float var = getVariance(i);
				float norm = var > 0 ? 1/var : 1; // No normalization if variance = 0;
				//if (m_NbAttrs < 15) System.out.println("  Normalization for: "+m_Attrs[i].getName()+" = "+norm);
				weights.setWeight(m_Attrs[i], norm);
			}
		}
	}
	
	/*
	 * Compute squared Euclidean distance between tuple's target attributes and this statistic's mean.
	 **/
	public float getSquaredDistance(DataTuple tuple, ClusAttributeWeights weights) {
		float sum = 0.0f;
		for (int i = 0; i < getNbAttributes(); i++) {
			NumericAttrType type = getAttribute(i);
			float dist = type.getNumeric(tuple) - m_Means[i];
			sum += dist * dist * weights.getWeight(type);
		}
		return sum / getNbAttributes();
	}
	
	public String getArrayOfStatistic(){
		NumberFormat fr = ClusFormat.SIX_AFTER_DOT;
		StringBuffer buf = new StringBuffer();
		buf.append("[");
		for (int i = 0; i < m_NbAttrs; i++) {
			if (i != 0) buf.append(",");
			buf.append(fr.format(m_Means[i]));
		}
		buf.append("]");
		return buf.toString();
	}

	public String getString(StatisticPrintInfo info) {
		NumberFormat fr = ClusFormat.SIX_AFTER_DOT;
		StringBuffer buf = new StringBuffer();
		buf.append("[");
		for (int i = 0; i < m_NbAttrs; i++) {
			if (i != 0) buf.append(",");
			float tot = getSumWeights(i);
			if (tot == 0) buf.append("?");
			else buf.append(fr.format(getSumValues(i)/tot));
		}
		buf.append("]");
		if (info.SHOW_EXAMPLE_COUNT_BYTARGET) {
			buf.append(": [");
			for (int i = 0; i < m_NbAttrs; i++) {
				if (i != 0) buf.append(",");
				buf.append(fr.format(m_SumWeights[i]));
			}
			buf.append("]");
		} else if (info.SHOW_EXAMPLE_COUNT) {
			buf.append(": ");
			buf.append(fr.format(m_SumWeight));
		}
		return buf.toString();
	}

	public String getPredictString() {
		StringBuffer buf = new StringBuffer();
		for (int i = 0; i < m_NbAttrs; i++) {
			if (i != 0) buf.append(",");
			buf.append(String.valueOf(m_Means[i]));
		}
		return buf.toString();
	}

	public String getDebugString() {
		NumberFormat fr = ClusFormat.THREE_AFTER_DOT;
		StringBuffer buf = new StringBuffer();
		buf.append("[");
		for (int i = 0; i < m_NbAttrs; i++) {
			if (i != 0) buf.append(",");
			buf.append(fr.format(getMean(i)));
		}
		buf.append("]");
		buf.append("[");
		for (int i = 0; i < m_NbAttrs; i++) {
			if (i != 0) buf.append(",");
			buf.append(fr.format(getVariance(i)));
		}
		buf.append("]");
		return buf.toString();
	}

	public void printDebug() {
		for (int i = 0; i < getNbAttributes(); i++) {
			float n_tot = m_SumWeight;
			float k_tot = m_SumWeights[i];
			float sv_tot = m_SumValues[i];
			float ss_tot = m_SumSqValues[i];
			System.out.println("n: "+n_tot+" k: "+k_tot);
			System.out.println("sv: "+sv_tot);
			System.out.println("ss: "+ss_tot);
			float mean = sv_tot / n_tot;
			float var = ss_tot - n_tot*mean*mean;
			System.out.println("mean: "+mean);
			System.out.println("var: "+var);
		}
		System.out.println("err: "+getError());
	}

	public void printDistribution(PrintWriter wrt) throws IOException {
		NumberFormat fr = ClusFormat.SIX_AFTER_DOT;
		for (int i = 0; i < m_Attrs.length; i++) {
			wrt.print(StringUtils.printStr(m_Attrs[i].getName(), 35));
			wrt.print(" [");
			wrt.print(fr.format(getMean(i)));
			wrt.print(",");
			wrt.print(fr.format(getVariance(i)));
			wrt.println("]");
		}
	}

	public void addPredictWriterSchema(String prefix, ClusSchema schema) {
		for (int i = 0; i < m_NbAttrs; i++) {
			ClusAttrType type = m_Attrs[i].cloneType();
			type.setName(prefix+"-p-"+type.getName());
			schema.addAttrType(type);
		}
	}

	public String getPredictWriterString() {
		StringBuffer buf = new StringBuffer();
		for (int i = 0; i < m_NbAttrs; i++) {
			if (i != 0) buf.append(",");
			if (m_Means != null) {
				buf.append(""+m_Means[i]);
			} else {
				buf.append("?");
			}
		}
		return buf.toString();
	}

	public void predictTuple(DataTuple prediction) {
		for (int i = 0; i < m_NbAttrs; i++) {
			NumericAttrType type = m_Attrs[i];
			type.setNumeric(prediction, m_Means[i]);
		}
	}

	public void vote(ArrayList votes) {
		reset();
		m_Means = new float[m_NbAttrs];
		RegressionStat vote;
		int nb_votes = votes.size();
		for (int j = 0; j < nb_votes; j++){
			vote = (RegressionStat) votes.get(j);
			for (int i = 0; i < m_NbAttrs; i++){
				m_Means[i] += vote.m_Means[i] / nb_votes;
			}
		}
	}

	public RegressionStat getRegressionStat() {
		return this;
	}
}
